/** A simple downloader for the flash contents for ch55x devices

    Since there is no direct download support the verify command is used to 
    download the contents. Unfortunatelly verify requires at least 16 bytes
    to do the compare and just one byte can be changed at a time.
    This makes the downloader very slow because its just fullspeed and the EPs
    have to be used in ping pong mode. The program will always download the 
    complete Codeflash and store it to an Intel hex file.
    
    It will use libusb for USB comunication. While it could work also with the 
    original VID / PID this not recommented because installing libusb on the WCH
    ids will prevent the WCHIspTool from working.

    Installation (WIN only):
    - flash the modified bootloader at startaddress 0
    - use zadig to install the libusbdriver (use libusbK not libusb on install)
    
    Modifications to the bootloader:
    - link to 0x000
    - changed device id to 0xFFFE
    - ask INT0 Pin for 0 to start the original bootloader (SW1 on Aarons board)
    - a small bugfix in USB setup handling (unstall EP0)  
*/

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <errno.h>
#include <string.h>
#include <conio.h>

#include "libusb.h"


struct libusb_device_handle *devh = NULL;


int ep_in_addr  = 0x82;
int ep_out_addr = 0x02;

uint32_t DefaultVendor = 0x4348;
uint32_t DefaultDevice = 0xFFFE;

uint16_t ProgSize = 0x8000;
uint8_t  ProgBuffer[0x8000];  //maximum programm size

int   Debuglevel = 0; 
char  fname[256] = "image.hex";

/** write a buffer to the device

    This code is taken from taken from librech.c
    \param [in] Buffer pointer to buffer to be send to the device
    \param [in] Size   size of this buffer
    \result            error code or bytes sent to the device    
*/
int write_to_device(uint8_t *Buffer, int Size)
{
   int actual_length;
   int rc = libusb_bulk_transfer( devh, 
                                  ep_out_addr, 
                                  Buffer, 
                                  Size,
                                 &actual_length, 4000); 
   if (rc < 0) 
   {
      fprintf(stderr, "Error while bulking out: %s\n",  libusb_error_name(rc));
      return rc;
   }
   return actual_length;
}

/** read a buffer from the device

    This code is taken from taken from librech.c
    \param [in] Buffer pointer to buffer for the read data
    \param [in] Size   size of this buffer
    \result            error code or number bytes valid in the buffer   
*/
int read_from_device(uint8_t *Buffer, int Size)
{
   int actual_length;
   int rc = libusb_bulk_transfer( devh, 
                                  ep_in_addr,
                                  Buffer,
                                  Size, 
                                 &actual_length,
                                  400000);
  if (rc == LIBUSB_ERROR_TIMEOUT) 
  {
    printf("timeout (%d)\n", actual_length);
    return -1;
  }
  else if (rc < 0)
  {
     fprintf(stderr, "Error while bulking in: %s\n",  libusb_error_name(rc));
     return rc;
  }
  return actual_length;
}

/** look for the device and claim it

    \param Vendor  vendor id to look for
    \param Product product id to look for
    \result        -1 on error or 0 if all ok
*/
int FindThisDevice(uint32_t Vendor,uint32_t Product)
{
   int rc = libusb_init(NULL);
   uint8_t buffer[100];
   if (rc < 0)
   {
      fprintf(stderr, "Error initializing libusb: %s\n", libusb_error_name(rc));
      return -1;
   }
   // Set debugging output to max level. */
   libusb_set_debug(NULL, Debuglevel);

   /* Look for a specific device and open it. */
   devh = libusb_open_device_with_vid_pid (NULL, Vendor, Product);
   if (!devh)
   {
      fprintf(stderr, "Can't find the requested device!\n");
      fprintf(stderr, "use zadig to connect a device to libusb!\n");
      return -1;
   }
   return libusb_claim_interface(devh, 0);
}

/** changes the scramble key to the desired value

    \param[in] Value new key 0 means scrample is off
    \result    error code or 0
*/
int SetScramble(uint32_t Value)
{
   uint8_t send[64];
   uint8_t recive[64];

   send[0] = 0xA6;
   send[1] = sizeof(Value);
   
   send[2] = Value >> 24 & 0xFF;
   send[3] = Value >> 16 & 0xFF;
   send[4] = Value >>  8 & 0xFF;
   send[5] = Value >>  0 & 0xFF;
   
   write_to_device(send,sizeof(send));
   
   int len = read_from_device(recive,sizeof(recive));
   if (len < 0) return len; //error
   return (int) recive[0];  
}

/** Connect the app to the bootloader

    first it asks for the loader version since currently just V1.1 is 
    supported. Then the WCH copyright string is requested which will 
    return the chip id.
    \return chip id or error code
*/
int ConnectToBootloader(void)
{
   uint8_t send[64];
   uint8_t recive[64];
   
   send[0] = 0xBB;
   write_to_device(send,sizeof(send));
   int len = read_from_device(recive,sizeof(recive));
   if (len < 0) return len; //error
   if ((len >0) && (recive[0]!=0x11))
   {
      fprintf(stderr, "wrong bootloader, just support for V1.1!\n");
      return -1;
   }
   send[0] = 0xA2;
   send[1] = 0x13;
   memcpy(&send[2],"USB DBG CH559 & ISP",0x14);
   write_to_device(send,sizeof(send));
   len = read_from_device(recive,sizeof(recive));
   if (len < 0) return len;             // error
   return (int) recive[0]; // chipid
}


void IsAlive(uint16_t Address)
{
   int static last;
   int current = Address >> 6;
   if (current != last)
   {
      fprintf(stderr, ".");
      last = current;
   }
}

/** look for a starting point for the brute force attack

    the function tries to find a byte sequence where 16 bytes filled with 0xFF. 
    If this point can be found the code address will be returned. The function 
    may take some time and even may fail if there is no point found. In that case
    there is no way to read back the Code Flash memory.       
    \param Address start address in CodeFlash for searching
    \result        found address or 0   
*/
int SearchStartAddress(uint16_t Start)
{
   uint8_t  send[64];
   uint8_t  recive[64];
   uint16_t addr = Start;

   SetScramble(0);

   fprintf(stderr, "    serching for 16 0xFF bytes  ");

   memset(send,0xFF,sizeof(send));
   send[0] = 0xA7;
   send[1] = 16;
   send[2] = addr >> 0 & 0xFF;
   send[3] = addr >> 8 & 0xFF;

   // try to find a point in Flashcode where 16 bytes are 0xff
   do
   {
      write_to_device(send,sizeof(send));
      read_from_device(recive,sizeof(recive));
      addr--;
      send[2] = addr >> 0 & 0xFF;
      send[3] = addr >> 8 & 0xFF;
      if (addr==0) break;
      IsAlive(addr);
   } while (recive[0]!=0);
   fprintf(stderr, "\n");
   return addr;
}


void BruteForceAttack(uint16_t Address, int dir)
{
   uint8_t  send[64];
   uint8_t  recive[64];
   uint16_t addr = Address;
   SetScramble(0);
   memset(send,0xFF,sizeof(send));
   if (dir == 0)
   {
      do
      {
         addr--;
         send[0] = 0xA7;
         send[1] = 16;
         send[2] = addr >> 0 & 0xFF;
         send[3] = addr >> 8 & 0xFF;
         int i=0;
         while(1) //this way 0x00 and 0xFF get prefered
         {        //much faster than other loops 
           send[4] = i;
           write_to_device(send,sizeof(send));
           read_from_device(recive,sizeof(recive));
           if(recive[0]==0) break;
           i--;
         }
         ProgBuffer[addr]= send[4];
         memmove(&send[5],&send[4],16);
         if (kbhit())
         {  //bored? ESC will cancel....
            if(getch()==27) return;
         }
         IsAlive(addr);
      }while (addr);
   }
   else
   {
      do
      {
         send[0] = 0xA7;
         send[1] = 16;
         send[2] = addr >> 0 & 0xFF;
         send[3] = addr >> 8 & 0xFF;
         for (int i=0; i<256; i++)
         {
            send[4+15] = i;
            write_to_device(send,sizeof(send));
            read_from_device(recive,sizeof(recive));
            if(recive[0]==0) break;
         }
         ProgBuffer[addr]= send[4];
         addr ++;
         memmove(&send[4],&send[5],16);
         if (kbhit())
         {  
            if(getch()==27) return;
         }
         IsAlive(addr);

      } while (addr < ProgSize-15);
      for (int i=0;i<16;i++) ProgBuffer[addr+i]= send[4+i];
   }     
}

/** not used at this time
*/
int LoadHexFile(uint8_t *filename)
{
   FILE     *file;
   uint8_t  line[80];
   uint8_t  tmp [5];
   uint8_t  nbytes,rec,b,chk;
   uint16_t addr;
   
   int      i;

   if (filename)
   {
      file = fopen(filename,"r");
      if (!file)
      {
         fprintf(stderr, "file not found!\n");
         return -1 ;
      }

      while (fgets(line,sizeof(line),file))
      {
         if (line[0] !=':')
         {
            fprintf(stderr, "not a intel hexfile!\n");
            return -2;
         }
         //fprintf(stderr,line);
         chk = 0;
         memcpy(tmp,&line[1],2);tmp[2]='\0';
         nbytes = (uint8_t) strtol(tmp,NULL,16);
         chk+= nbytes;

         memcpy(tmp,&line[3],4);tmp[4]='\0';
         addr   = (uint16_t) strtol(tmp,NULL,16);
         chk += addr>>8; chk += addr & 0xFF;

         memcpy(tmp,&line[7],2);tmp[2]='\0';
         rec    = (uint8_t) strtol(tmp,NULL,16);
         chk += rec;

         if (rec==1) break;

         // todo check for overflow
         if (addr+nbytes > sizeof(ProgBuffer))
         {
            fprintf(stderr, "to many bytes in Hex file!\n");
            return -3;
         }
         for (i=0;i< nbytes;i++)
         {
            memcpy(tmp,&line[i*2+9],2);tmp[2]='\0';
            b = (uint8_t) strtol(tmp,NULL,16);
            chk +=b;
            ProgBuffer[addr+i]= b;
         }
         chk = (~chk)+1;
         memcpy(tmp,&line[i*2+9],2);tmp[2]='\0';
         b = (uint8_t) strtol(tmp,NULL,16);
         if (b != chk)
         {
            fprintf(stderr, "intel hex checksum error!\n");
            return -4;
         }
      }
      fclose(file);
      fprintf(stderr, "hexfile loaded!\n");
      return 0;
   }
   return -1;
}

/**  store the contents to a Intel hex file
*/
int WriteIntelHex(uint8_t *filename)
{
   FILE     *file;
   uint8_t  line[80];
   uint8_t  tmp [5];
   uint8_t  nbytes,b,chk;
   uint16_t addr;
   
   int      i;

   if (filename)
   {
      file = fopen(filename,"wt");
      if (!file)
      {
         fprintf(stderr, "creating hexfile failed!\n");
         return -1 ;
      }
      addr=0;
      while (addr < ProgSize)
      {
      	 line[0]= ':';
      	 line[1]= '1';
      	 line[2]= '0';      	 
      	 chk= 16;
      	 sprintf(&line[3],"%04X",addr); chk+= addr & 0xFF;chk+= addr >>8 & 0xFF;
      	 line[7]='0'; line[8]='0';chk+=0;
      	 for (i=0;i<16;i++)
      	 {
      	 	  sprintf(&line[i*2+9],"%02X",ProgBuffer[addr+i]);
      	 	  chk+= ProgBuffer[addr+i];
      	 }
      	 sprintf(&line[i*2+9],"%02X",chk); 
      	 line[i*2+11]='\0';
      	 fprintf (file,"%s\n",line);
      	 addr +=16;
      }
      fprintf (file,":00000001FF\n");
      fclose(file);
      return 0;
   }
   return -1;
}

void Download(void)
{
   int result;
   fprintf(stderr, "starting download ...\n");
    
   if (FindThisDevice(DefaultVendor,DefaultDevice)==0)
   {
      fprintf(stderr, "    device found ready to query!\n");
   }
   else return;
   result = ConnectToBootloader();
   if (result < 0) return;
   switch (result)
   {
      case 0x51:
           fprintf(stderr, "    CH551 chip detected!\n");
           ProgSize = 0x3000;
           break; 
      case 0x52:
           fprintf(stderr, "    CH552 chip detected!\n");
           ProgSize = 0x4000;
           break;
      case 0x53:
           fprintf(stderr, "    CH553 chip detected!\n");
           ProgSize = 0x3000;
           break; 
      case 0x54:
           fprintf(stderr, "    CH554 chip detected!\n");
           ProgSize = 0x4000;
           break; 

      default:
           fprintf(stderr, "    unknown chip detected!\n");
           return;
   }

   memset (ProgBuffer,0xFF,ProgSize);
   uint16_t Address = SearchStartAddress(ProgSize-0x10);
   if (Address==0)
   {
      fprintf(stderr, "no address found! Can't continue!\n");
      return;
   }
   else
   {
      fprintf(stderr, "    found a starting point for reading the chip\n");
      fprintf(stderr, "    address : 0x%04X\n",Address);
   }  

   BruteForceAttack(Address,0);  //going down til 0x0000
   BruteForceAttack(Address,1);  //      up   til ProgSize reached

   if (devh)
   {
      libusb_release_interface(devh, 0);
      libusb_close(devh);
   }
   WriteIntelHex(fname);
}

void DisplayHelp(void)
{
   printf("\noptions:\n");
   printf("    -v[xxxx] : Vendor Id to look for       (default: -v0x4348)\n");
   printf("    -d[xxxx] : Device Id to look for       (default: -d0xFFFE)\n");
   printf("    -fNAME   : Filename Intel Hex file     (default: -fimage.hex)\n");
   printf("    -x[0..3] : Libusb errorlevel           (default: -x0)\n");
   printf("\n");
   printf("examples : downloader -v0xABCD      set vendor to 0xABCD\n");
   printf("           downloader -fmyfile.hex  save hex in myfile.hex\n");
}

void main(int argc, char *argv[])
{
   
   printf("\CH55x firmware downloader \n");
   /* collect the options */
   for (int i=1;i < argc; i++)
   {
      if ((*argv[i]=='-') || (*argv[i]=='/'))
      {  /* all params start with - or / */
         argv[i]++;
         char option = toupper (*argv[i]++);
         switch (option)
         {
            case 'D':
                 DefaultDevice=strtol (argv[i],NULL,0);
                 break;
            case 'F':
                 strcpy(fname,argv[i]);
                 break;
            case 'V':
                 DefaultVendor=strtol (argv[i],NULL,0);
                 break;
            case 'X':
                 Debuglevel = strtol (argv[i],NULL,0);
                 if(Debuglevel>3) Debuglevel=3; 
                 break;
            default:  
                 DisplayHelp();
                 exit(0);
         }
      }
   }   
   printf ("no parameter given using the defaults!\n");
   Download();
}   